



rd_depression_filling --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
rd_breach_depressions --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
rd_flow_accumulation --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
rd_terrain_attribute --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
rd_info --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
rd_compare --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
